package p4i;
#
#Perl implementation of the UPU Postal-4i symbology
#
#Copyright (C) 2006 J. Wells
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without modification,
#are permitted provided that the following conditions are met:
#1. Redistributions of source code shall retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#2. Redistributions in binary form shall reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#3. The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#IMPORTANT - see key notes introduced by >>>>
#
#
#This implementation is intended to provide an easy to follow illustration of the
#basic Postal-4i generation algorithm. It is designed for transparency, not efficiency!
#
#usage
	#there are three externally callable subroutines
	#$p4icode = p4i::gen($formatcode, $leftsync, $rightsync, $binstring, $cvalue, $codelen);
					#returns postal-4i bar code in FADT format for the given parameters
	#p4i::print4s ($p4icode, $barheight, $barpitch, $ocrdata, $binstring);
					#prints the barcode $p4icode with subtitles $ocrdata and $binstring
	#p4i::closeup		#closes up the xml file created by print4s

use Reedsolomon;		#assumes package rs is in file Reedsolomon.pm

sub gen {			#generate postal-4i bar code from supplied parameters

	#look up table for conversion of decimal codeword values to 3-bar groups
	my @dectobars = ("FFF", "FFA", "FFD", "FFT", "FAF", "FAA", "FAD", "FAT",
			"FDF", "FDA", "FDD", "FDT", "FTF", "FTA", "FTD", "FTT",
			"AFF", "AFA", "AFD", "AFT", "AAF", "AAA", "AAD", "AAT",
			"ADF", "ADA", "ADD", "ADT", "ATF", "ATA", "ATD", "ATT",
			"DFF", "DFA", "DFD", "DFT", "DAF", "DAA", "DAD", "DAT",
			"DDF", "DDA", "DDD", "DDT", "DTF", "DTA", "DTD", "DTT",
			"TFF", "TFA", "TFD", "TFT", "TAF", "TAA", "TAD", "TAT",
			"TDF", "TDA", "TDD", "TDT", "TTF", "TTA", "TTD", "TTT");

	#pick up calling parameters
	my $formatcode = shift;				#format code to be used
	my $leftsync = shift;				#left sync value
	my $rightsync = shift;				#right sync value
	my $binstring = shift;				#binary string to be represented in bar code
	my $cvalue = shift;					#no of variable codewords parameter
	my $codelen = shift;					#length of final bar code

	#set up polynomial coefficients
	$binstring = $formatcode.$binstring;		#prefix binary string with format code
	substr($binstring,-12,0) = $rightsync;	#insert rightsync 12 bits from the end

	#insert leftsync 12 bits from beginning if format code starts with 00
	if ($formatcode =~ /^00/) {substr($binstring,12,0) = $leftsync}

	my @c = (0) x 12;					#c0 to c11 all 0

	#c12 until all string elements used up
	while ($binstring ne "") {push @c,bintodec(substr($binstring,-6,6,""))}

	#rest up to c61 are either 0 or leftsync
	while (@c < 62) {($formatcode =~ /^00/) ? (push @c,0) : (push @c,bintodec($leftsync))}

	#c62 is either 0 or formatcode
	($formatcode =~ /^00/) ? (push @c,0) : (push @c,bintodec($formatcode));

#print "polynomial @c\n";		#remove hash at beginning to print polynomial coefficients

	my @r = rs::encode(\@c);			#calculate reed solomon error correction codewords
	
#print "ecc @r\n";					#remove hash at beginning to print rs ecc codewords

	#extract codewords to be printed
	my @bcode;
	unshift @bcode,$dectobars[$c[12]];	#c12 converted to bars
	unshift @bcode,$dectobars[$c[13]];	#c13 converted to bars
	unshift @bcode,$dectobars[$c[14]];	#c14 converted to bars

	if ($c[62] ge 16) {				#if format code was 6 bits
		$index = 11;
		if ($codelen < 48) {$index = int($codelen/3) - 5}
		while ($index ge 0) {unshift @bcode,$dectobars[$r[11-$index--]]}

		$index = int($codelen/3) - 15;
		if ($index < 1) {$index = 1}
		while ($index > 0) {unshift @bcode,$dectobars[$c[$cvalue+12-$index--]]}
	}
	else {						#if format code was only 4 bits (ID-tag)
		$index = 11;
		if ($codelen < 3*$cvalue + 36) {$index = int($codelen/3) - $cvalue - 1}
		while ($index ge 0) {unshift @bcode,$dectobars[$r[11-$index--]]}

		$index = $cvalue - 3;
		while ($index > 0) {unshift @bcode,$dectobars[$c[$cvalue+12-$index--]]}
	}

	return join("",@bcode);				#return the result
}

sub print4s {					#print postal-4i bar code at given height and pitch
#
#A practical implementation would probably make use of a 4-state font for the printing
#of the bars. This example implementation is designed to run on any machine which
#supports Perl and therefore cannot make use of a font definition, since the machine
#might not have the required font installed. It instead prints the bar code to an xml
#file which can be opened by Microsoft Word. This makes the print4s subroutine, which
#includes needed header and trailer information, rather messy!
#
#The print routine generates the necessary xml to print the postal-4i bar code followed by
#a supplied sub-title, normally intended to be the human readable equivalent of the data
#represented in the bar code. Allowance is also made for printing of the character (FADT)
#based equivalent of the bar code, but this is disabled in the following (hash in front
#of the print statement, converting it to a comment).
#
#To permit the generation and printing of multiple Postal-4i 
#bar codes, print4s does not output the xml trailer for the file. The calling programme has
#to explicitly ask for this by using p4i::closeup
#
	#>>>>creates an xml output file that can be opened with microsoft word
	#>>>>file is left open at end; it must be closed by using p4i::closeup

	my $bardata = shift;			#data as FADT string
	my $bh = shift;				#height in cm
	my $bp = shift;				#pitch in cm
	my $title1 = shift;			#1st subtitle
	my $nobars = length($bardata);	#no of bars
	my $bl = $nobars * $bp;		#overall length = pitch * no of bars

	unless (defined $fileopen) {	#first time we need to open & initialise output file

		%symbols = (
			header => "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?><?mso-application progid=\"Word.Document\"?><w:wordDocument xmlns:w=\"http://schemas.microsoft.com/office/word/2003/wordml\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:sl=\"http://schemas.microsoft.com/schemaLibrary/2003/core\" xmlns:aml=\"http://schemas.microsoft.com/aml/2001/core\" xmlns:wx=\"http://schemas.microsoft.com/office/word/2003/auxHint\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:dt=\"uuid:C2F41010-65B3-11d1-A29F-00AA00C14882\" w:macrosPresent=\"no\" w:embeddedObjPresent=\"no\" w:ocxPresent=\"no\" xml:space=\"preserve\"><o:DocumentProperties><o:Title>UPU Postal-4i bar code example</o:Title><o:Author>UPU S48</o:Author></o:DocumentProperties><w:styles><w:versionOfBuiltInStylenames w:val=\"4\"/><w:latentStyles w:defLockedState=\"off\" w:latentStyleCount=\"156\"/><w:style w:type=\"paragraph\" w:default=\"on\" w:styleId=\"Normal\"><w:name w:val=\"Normal\"/><w:rsid w:val=\"00237A94\"/><w:pPr><w:jc w:val=\"both\"/></w:pPr><w:rPr><w:rFonts w:ascii=\"Arial\" w:h-ansi=\"Arial\"/><wx:font wx:val=\"Arial\"/><w:sz-cs w:val=\"24\"/><w:lang w:val=\"EN-GB\" w:fareast=\"EN-US\" w:bidi=\"AR-SA\"/></w:rPr></w:style><w:style w:type=\"character\" w:default=\"on\" w:styleId=\"DefaultParagraphFont\"><w:name w:val=\"Default Paragraph Font\"/><w:semiHidden/></w:style></w:styles><w:body><wx:sect>",
			begin => "<w:p><w:r><w:pict><v:group id=\"_x0000_s1027\" editas=\"canvas\" style=\"width:#lcm;height:#hcm;mso-position-horizontal-relative:char;mso-position-vertical-relative:line\" coordorigin=\"0,0\" coordsize=\"#n,3\">",
			F => "<v:line id=\"F\" style=\"position:absolute\" from=\"#b,0\" to=\"#b,3\" strokeweight=\".04cm\"/>",
			A => "<v:line id=\"A\" style=\"position:absolute\" from=\"#b,0\" to=\"#b,2\" strokeweight=\".04cm\"/>",
			D => "<v:line id=\"D\" style=\"position:absolute\" from=\"#b,1\" to=\"#b,3\" strokeweight=\".04cm\"/>",
			T => "<v:line id=\"T\" style=\"position:absolute\" from=\"#b,1\" to=\"#b,2\" strokeweight=\".04cm\"/>",
			end => "<w10:wrap type=\"none\"/><w10:anchorlock/></v:group></w:pict></w:r>",
			ocr => "<w:r><w:br/></w:r><w:r><w:t>#d</w:t></w:r></w:p>",
			trailer => "<w:sectPr><w:pgSz w:w=\"11907\" w:h=\"16840\" w:code=\"9\"/><w:pgMar w:top=\"1418\" w:right=\"1418\" w:bottom=\"1418\" w:left=\"1418\" w:header=\"709\" w:footer=\"709\" w:gutter=\"0\"/><w:cols w:space=\"708\"/><w:docGrid w:line-pitch=\"360\"/></w:sectPr></wx:sect></w:body></w:wordDocument>"
		);

		open (BARCODE,">","postal-4i barcode.xml") or die "Can't open output file: $!";
		print BARCODE "$symbols{\"header\"}";		#common header data
		$fileopen = 1;
	}

	my $begindata = $symbols{"begin"};				#begin data for a bar code
	$begindata =~ s/#h/$bh/;						#substitute bar height
	$begindata =~ s/#l/$bl/;						#substitute overall length
	$begindata =~ s/#n/$nobars/;					#and number of bars
	print BARCODE "$begindata";					#create picture frame

	for (my $i=0; $i<$nobars; $i++) {				#for each bar
		my $bar = $symbols{substr($bardata,$i,1)};	#pick up bar definition
		$bar =~ s/#b/$i/g;						#insert its horizontal position
		print BARCODE "$bar";					#output to file
	}
	print BARCODE "$symbols{\"end\"}";				#complete the picture
	print BARCODE "<w:r><w:br/></w:r><w:r><w:t>$title1</w:t></w:r></w:p>";	#add title
#	print BARCODE "<w:p><w:r><w:t>$bardata</w:t></w:r></w:p>";			#add title2
	print BARCODE "<w:p/><w:p/>";					#terminate with blank lines
}

sub bintodec {				#crude binary to decimal converter
	my $bin = shift;		#value to convert
	my $dec = 0;
	while (length($bin) > 0) {$dec = 2*$dec + substr($bin,0,1); $bin = substr($bin,1)}
	return $dec;
}

sub closeup {
	if (defined $fileopen) {					#finish up and close output file
			print BARCODE "$symbols{\"trailer\"}";
			close BARCODE;
			undef $fileopen;

	}
}

1;	#so use succeeds if the package is put in a separate .pm file
