package tag4i;
#
#Perl implementation of the UPU 4-state ID-tag
#
#Copyright (C) 2006 J. Wells
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without modification,
#are permitted provided that the following conditions are met:
#1. Redistributions of source code shall retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#2. Redistributions in binary form shall reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#3. The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#This implementation is intended to provide an easy to follow illustration of the
#generation of Postal-4i ID-tags. It is designed for transparency, not efficiency!
#
#usage
	#tag4i::gen($idtag, $length) where:
	#$idtag is the message representation of the ID-tag to be encoded
	#$length is the desired length of the resulting bar code (57 or 75 bars)

use Postal4i; 						#assumes package p4i is in file Postal4i.pm

sub d2b {										#crude decimal to binary converter
	my $dec = shift;						#value to convert
	my $len = shift;						#minimum length required
	#NB if length = 0 returns minimum length binary string, with top (left) bit = 1
	my $bin = "";
	while ($dec > 0) {
		$bin = ($dec%2).$bin;
		$dec = int($dec/2);
	}
	while (length($bin) < $len) {$bin = "0".$bin}	#if shorter than needed, pad 0's
	return $bin;
}

sub verifyinput {						#verify that supplied data are valid

	undef $formatcode;							#assume invalid
	if ($idtag =~ /^J18C([A-Z][A-Z0-9]{2})([A-F0-9]{3})([NHLU])(\d\d)(\d\d)(\d\d)([0-5])(\d\d\d\d\d)([DFNT])$/) {						#J18C format
		$formatcode = "0010";
		$tagiss = $1;
		$tageq = $2;
		$tagpri = $3;
		$tagmon = $4;
		$tagday = $5;
		$taghour = $6;
		$tagmin = $7;
		$tagserial = $8;
		$tagtrack = $9;
		if ($tagmon < 1) {undef $formatcode}		#validate month, day, hour and serial
		if ($tagmon > 12) {undef $formatcode}
		if ($tagday < 1) {undef $formatcode}
		if ($tagday > 31) {undef $formatcode}
		if ($taghour > 23) {undef $formatcode}
		if ($tagserial > 16383) {undef $formatcode}
	}
	elsif ($idtag =~ /^J18D([A-Z][A-Z0-9]{2})([A-F0-9]{3})([NHLU])(\d{12})([DFNT])$/) {
		$formatcode = "0011";						#J18D format
		$tagiss = $1;
		$tageq = $2;
		$tagpri = $3;
		$tagserial = $4;
		$tagtrack = $5;
		if ($tagserial > 1073741823) {undef $formatcode}	#validate serial number
	}
	unless (defined $formatcode) {					#anything else is illegal
		print "ID-tag $idtag invalid; please re-enter\n";
		print "J18Ciiieeepmmddhhfssssst or J18Diiieeepsssssssssssst format\n";
		$idtag = <STDIN>; chomp $idtag; verifyinput();
	}

	unless ($length eq 57) {unless ($length eq 75) {	#length must be 57 or 75
		print "bar code length $length invalid; please re-enter\n";
		$length = <STDIN>; chomp $length; verifyinput();
	}}
}

sub calcstring {					#create input data string from supplied data
	%cc = ("0" => 35, "1" => 34, "2" => 33, "3" => 32, "4" => 31, "5" => 30, 
		"6" => 29, "7" => 28,  "8" => 27, "9" => 26, "A" => 25, "B" => 24, 
		"C" => 23, "D" => 22, "E" => 21, "F" => 20, "G" => 19, "H" => 18, 
		"I" => 17, "J" => 16, "K" => 15, "L" => 14, "M" => 13, "N" => 12, 
		"O" => 11, "P" => 10, "Q" => 9, "R" => 8, "S" => 7, "T" => 6, 
		"U" => 5, "V" => 4, "W" => 3, "X" => 2, "Y" => 1, "Z" => 0);
	%hc = ("0" => "0000", "1" => "0001", "2" => "0010", "3" => "0011", 
		"4" => "0100", "5" => "0101", "6" => "0110", "7" => "0111",  
		"8" => "1000", "9" => "1001", "A" => "1010", "B" => "1011", 
		"C" => "1100", "D" => "1101", "E" => "1110", "F" => "1111");
	%pc = ("N" => "00", "L" => "01", "H" => "10", "U" => "11");
	%tc = ("T" => "00", "F" => "01", "D" => "10", "N" => "11");

	#bit numbers here relate to position in the ID-tag binary representation
	$b04_19 = d2b(1600*$cc{substr($tagiss,0,1)} +40*$cc{substr($tagiss,1,1)} + 
					$cc{substr($tagiss,2,1)},16);
	$b20_31 = $hc{substr($tageq,0,1)}.$hc{substr($tageq,1,1)}.$hc{substr($tageq,2,1)};
	$b32_33 = $pc{$tagpri};
	if ($formatcode eq "0011") {$b34_63 = d2b($tagserial,30)}
	else {$b34_63 = d2b(16384*(5120*($tagmon-1) + 160*$tagday + 6*$taghour+$tagmin) + 
					$tagserial,30)}
	return $b04_19.$b20_31.$b32_33.substr($b34_63,0,20).$tc{$tagtrack}.substr($b34_63,20);
}

sub gen {

	$leftsync = "010110";				#l is 22 
	$rightsync = "100110";				#38 decimal
	$cvalue = 13;						#not actually used
	$barheight = 0.5;					#bar height in cm
	$barpitch = 0.115;					#pitch in cm

	$idtag = shift;						#ID-tag
	$length = shift;					#length

	verifyinput();
	$ocrdata = "(J) ".$idtag;

	$binstring = calcstring();				#calculate input string, generate and print
	$p4icode = p4i::gen($formatcode, $leftsync, $rightsync, $binstring, $cvalue, $length);
	p4i::print4s($p4icode, $barheight, $barpitch, $ocrdata);
}

1;	#so use succeeds if the package is put in a separate .pm file
